(** * Problem 1 *)

(** ** Part I: The definitions you'll need to work with *)

(** A basic binary tree type, with no data values, just tree structure *)
Inductive tree :=
| Leaf
| Node (t1 t2 : tree).

(** How many [Node]s were used to build a tree? *)
(* (It may be helpful to review the definition of the natural numbers [nat]
 * from class, since Coq's standard library uses the same definition. *)
Fixpoint size (t : tree) : nat :=
  match t with
    | Leaf => O
    | Node t1 t2 => S (size t1 + size t2)
  end.

(** A silly function that grows a tree by duplication *)
Fixpoint grower (t : tree) : tree :=
  match t with
    | Leaf => Leaf
    | Node t1 t2 => Node t1 (Node t2 (grower t2))
  end.


(** ** Part II: Some arithmetic lemmas you may find useful *)

Require Import Omega.

Lemma leRefl : forall n,
  n <= n.
Proof.
  intros; omega.
Qed.

Lemma le_S : forall n m,
  n <= m
  -> n <= S m.
Proof.
  intros; omega.
Qed.

Lemma leSS : forall n m,
  n <= m
  -> S n <= S m.
Proof.
  intros; omega.
Qed.

Lemma leCong : forall a b c d,
  a <= c
  -> b <= d
  -> a + b <= c + d.
Proof.
  intros; omega.
Qed.

Lemma leTrans : forall a b c,
  a <= b
  -> b <= c
  -> a <= c.
Proof.
  intros; omega.
Qed.

Lemma lePlus1 : forall a b,
  a <= a + b.
Proof.
  intros; omega.
Qed.

Lemma lePlus2 : forall a b,
  a <= b + a.
Proof.
  intros; omega.
Qed.


(** * Problem 2 *)

Require Import String.

(** ** Part I: Definitions translated from the Problem Set 2 PDF *)

(** Synonym for variable names *)
Definition var := string.

(** Syntax of lambda terms with exceptions *)
Inductive exp :=
| Var (x : var)
| App (e1 e2 : exp)
| Abs (x : string) (e1 : exp)
| Throw (e1 : exp)
| TryCatch (e1 : exp) (x : var) (e2 : exp).

(** Substitution in terms (without the complexities of capture avoidance) *)
Fixpoint subst (e e' : exp) (x : var) : exp :=
  match e with
    | Var y => if string_dec y x then e' else Var y
    | App e1 e2 => App (subst e1 e' x) (subst e2 e' x)
    | Abs y e1 => Abs y (if string_dec y x then e1 else subst e1 e' x)
    | Throw e1 => Throw (subst e1 e' x)
    | TryCatch e1 y e2 => TryCatch (subst e1 e' x) y (if string_dec y x then e2 else subst e2 e' x)
  end.

(** Final results of big-step evaluation *)
Inductive result :=
| Normal (e : exp)
| Exception (e : exp).

(** The big-step evaluation relation *)
Inductive bigStep : exp -> result -> Prop :=
| BigAbs : forall x e1,
  bigStep (Abs x e1) (Normal (Abs x e1))
| BigApp : forall e1 e2 x e1' v r,
  bigStep e1 (Normal (Abs x e1'))
  -> bigStep e2 (Normal v)
  -> bigStep (subst e1' v x) r
  -> bigStep (App e1 e2) r

| BigAppExn1 : forall e1 e2 v,
  bigStep e1 (Exception v)
  -> bigStep (App e1 e2) (Exception v)
| BigAppExn2 : forall e1 e2 x e1' v,
  bigStep e1 (Normal (Abs x e1'))
  -> bigStep e2 (Exception v)
  -> bigStep (App e1 e2) (Exception v)

| BigThrow : forall e1 v,
  bigStep e1 (Normal v)
  -> bigStep (Throw e1) (Exception v)
| BigThrowExn : forall e1 v,
  bigStep e1 (Exception v)
  -> bigStep (Throw e1) (Exception v)

| BigTryCatch : forall e1 x e2 v,
  bigStep e1 (Normal v)
  -> bigStep (TryCatch e1 x e2) (Normal v)
| BigTryCatchExn : forall e1 x e2 v r,
  bigStep e1 (Exception v)
  -> bigStep (subst e2 v x) r
  -> bigStep (TryCatch e1 x e2) r.

(** Evaluation contexts *)
Inductive context :=
| Hole
| App1 (C : context) (e : exp)
| App2 (v : exp) (C : context)
| Throw1 (C : context)
| TryCatch1 (C : context) (x : var) (e : exp).

(** Restricted evaluation contexts, preventing descent beneath [TryCatch] *)
Inductive contextNoCatch :=
| Hole'
| App1' (C : contextNoCatch) (e : exp)
| App2' (v : exp) (C : contextNoCatch)
| Throw1' (C : contextNoCatch).

(** How to plug an expression into a normal context *)
Fixpoint plug (C : context) (e : exp) : exp :=
  match C with
    | Hole => e
    | App1 C e' => App (plug C e) e'
    | App2 v C => App v (plug C e)
    | Throw1 C => Throw (plug C e)
    | TryCatch1 C x e' => TryCatch (plug C e) x e'
  end.

(** How to plug an expression into a restricted context *)
Fixpoint plugNoCatch (C : contextNoCatch) (e : exp) : exp :=
  match C with
    | Hole' => e
    | App1' C e' => App (plugNoCatch C e) e'
    | App2' v C => App v (plugNoCatch C e)
    | Throw1' C => Throw (plugNoCatch C e)
  end.

(** When is an expression a value? *)
Inductive isValue : exp -> Prop :=
| ValueAbs : forall x e,
  isValue (Abs x e).

(** One step of the small-step, contextual semantics *)
Inductive contextStep : exp -> exp -> Prop :=
| ContextBeta : forall C x e v,
  isValue v
  -> contextStep (plug C (App (Abs x e) v)) (plug C (subst e v x))
| ContextTryCatch : forall C v x e,
  isValue v
  -> contextStep (plug C (TryCatch v x e)) (plug C v)
| ContextTryCatchExn : forall C v x e,
  isValue v
  -> contextStep (plug C (TryCatch (Throw v) x e)) (plug C (subst e v x))
| ContextUncaught : forall C Cn v,
  isValue v
  -> contextStep (plug C (plugNoCatch Cn (Throw v))) (plug C (Throw v)).

(** NOTE FOR COQ 8.3 USERS:
  * You'll probably want to [eapply] the following lemma instead of [ContextUncaught]. *)

Lemma ContextUncaught' : forall C Cn v e1 e2,
  e1 = plug C (plugNoCatch Cn (Throw v))
  -> e2 = plug C (Throw v)
  -> isValue v
  -> contextStep e1 e2.
Proof.
  intros; subst; apply ContextUncaught; assumption.
Qed.

(** Transitive-reflexive closure of [contextStep] *)
Inductive contextStepStar : exp -> exp -> Prop :=
| StarRefl : forall e,
  contextStepStar e e
| StarStep : forall e e' e'',
  contextStep e e'
  -> contextStepStar e' e''
  -> contextStepStar e e''.

(** Conversion of big-step results to expressions more suited to small-step *)
Definition resultToExp (r : result) : exp :=
  match r with
    | Normal v => v
    | Exception v => Throw v
  end.


(** ** Part II: Some definitions and lemmas that we use in our solution *)
(* (You aren't required to use any of these, but most will probably be useful in any case.) *)

(** Many-step evaluation is transitive. *)
Lemma contextStepStar_trans : forall e e' e'',
  contextStepStar e e'
  -> contextStepStar e' e''
  -> contextStepStar e e''.
Proof.
  induction 1; intros.

  assumption.

  eapply StarStep.
  eassumption.
  apply IHcontextStepStar.
  assumption.
Qed.

(** How to plug one context ([C2]) into the hole found in another ([C1]) *)
Fixpoint compose (C1 C2 : context) : context :=
  match C1 with
    | Hole => C2
    | App1 C1 e => App1 (compose C1 C2) e
    | App2 v C1 => App2 v (compose C1 C2)
    | Throw1 C1 => Throw1 (compose C1 C2)
    | TryCatch1 C1 x e => TryCatch1 (compose C1 C2) x e
  end.

(** Composition works as expected with [plug]. *)
Lemma compose_correct : forall C1 C2 e,
  plug C1 (plug C2 e) = plug (compose C1 C2) e.
Proof.
  induction C1; simpl; intros.

  reflexivity.

  rewrite IHC1.
  reflexivity.

  rewrite IHC1.
  reflexivity.

  rewrite IHC1.
  reflexivity.

  rewrite IHC1.
  reflexivity.
Qed.

(** An extra evaluation context may be "wrapped around" any single step. *)
Lemma descend : forall C e e',
  contextStep e e'
  -> contextStep (plug C e) (plug C e').
Proof.
  destruct 1; intros.

  rewrite compose_correct.
  rewrite compose_correct.
  apply ContextBeta.
  assumption.

  rewrite compose_correct.
  rewrite compose_correct.
  apply ContextTryCatch.
  assumption.

  rewrite compose_correct.
  rewrite compose_correct.
  apply ContextTryCatchExn.
  assumption.

  rewrite compose_correct.
  rewrite compose_correct.
  apply ContextUncaught.
  assumption.
Qed.

(** Like in the lecture demo code, we prove an alternative version, which makes for nicer proofs
  * in Coq 8.3.  Coq 8.4 can make do with just the original.
  * Our solution actually doesn't invoke either of these directly, so you can probably ignore
  * this wrinkle! *)
Lemma descend' : forall C e e' e1 e2,
  contextStep e e'
  -> e1 = plug C e
  -> e2 = plug C e'
  -> contextStep e1 e2.
Proof.
  intros; subst; apply descend; assumption.
Qed.

(** *** Next, a series of lemmas for pushing [contextStepStar] judgments "inside" more syntax *)

Theorem stepApp1 : forall e1 e2 e1',
  contextStepStar e1 e1'
  -> contextStepStar (App e1 e2) (App e1' e2).
Proof.
  induction 1.

  apply StarRefl.

  eapply StarStep.
  eapply (descend' (App1 Hole e2)).
  eassumption.
  reflexivity.
  reflexivity.
  simpl.
  assumption.
Qed.

Theorem stepApp2 : forall e1 e2 e2',
  contextStepStar e2 e2'
  -> isValue e1
  -> contextStepStar (App e1 e2) (App e1 e2').
Proof.
  induction 1; intros.

  apply StarRefl.

  eapply StarStep.
  eapply (descend' (App2 e1 Hole)).
  eassumption.
  reflexivity.
  reflexivity.
  simpl.
  apply IHcontextStepStar.
  assumption.
Qed.

Theorem stepThrow : forall e e',
  contextStepStar e e'
  -> contextStepStar (Throw e) (Throw e').
Proof.
  induction 1; intros.

  apply StarRefl.

  eapply StarStep.
  eapply (descend' (Throw1 Hole)).
  eassumption.
  reflexivity.
  reflexivity.
  simpl.
  apply IHcontextStepStar.
Qed.

Theorem stepTryCatch : forall e1 x e2 e1',
  contextStepStar e1 e1'
  -> contextStepStar (TryCatch e1 x e2) (TryCatch e1' x e2).
Proof.
  induction 1; intros.

  apply StarRefl.

  eapply StarStep.
  eapply (descend' (TryCatch1 Hole x e2)).
  eassumption.
  reflexivity.
  reflexivity.
  simpl.
  apply IHcontextStepStar.
Qed.

(** When is a [result] a value? *)
Inductive resultValue : result -> Prop :=
| RNormal : forall v,
  isValue v
  -> resultValue (Normal v)
| RException : forall v,
  isValue v
  -> resultValue (Exception v).

(** A big-step result is always a value. *)
Theorem bigValue : forall e1 r,
  bigStep e1 r
  -> resultValue r.
Proof.
  induction 1.

  apply RNormal.
  apply ValueAbs.

  assumption.

  assumption.

  assumption.

  inversion_clear IHbigStep.
  apply RException.
  assumption.

  assumption.

  assumption.

  assumption.
Qed.
